#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013 Artem Glebov <artem.glebov@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import GLib, GObject

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.display.name import displayer as name_displayer

# -------------------------------------------------------------------------
#
# grabbers constants and routines
#
# -------------------------------------------------------------------------
from .grabbers import (
    grabber_generators,
    can_grab,
    grabber_position,
    switch_grabber,
    CURSORS,
    GRABBER_INSIDE,
    INSIDE,
    INNER_GRABBERS,
    OUTER_GRABBERS,
    MOTION_FUNCTIONS,
)

# -------------------------------------------------------------------------
#
# PhotoTaggingGramplet
#
# -------------------------------------------------------------------------

RESIZE_RATIO = 1.5
MAX_ZOOM = 10
MIN_ZOOM = 0.05
MAX_SIZE = 2000
MIN_SIZE = 50
SHADING_OPACITY = 0.7
MIN_SELECTION_SIZE = 10


def scale_to_fit(orig_x, orig_y, target_x, target_y):
    """
    Calculates the scale factor to fit the rectangle
    orig_x * orig_y by scaling keeping the aspect ratio.
    """
    orig_aspect = orig_x / orig_y
    target_aspect = target_x / target_y
    if orig_aspect > target_aspect:
        return target_x / orig_x
    else:
        return target_y / orig_y


def resize_keep_aspect(orig_x, orig_y, target_x, target_y):
    """
    Calculates the dimensions of the rectangle obtained from
    the rectangle orig_x * orig_y by scaling to fit
    target_x * target_y keeping the aspect ratio.
    """
    orig_aspect = orig_x / orig_y
    target_aspect = target_x / target_y
    if orig_aspect > target_aspect:
        return (target_x, target_x * orig_y // orig_x)
    else:
        return (target_y * orig_x // orig_y, target_y)


def order_coordinates(point1, point2):
    """
    Returns the rectangle (x1, y1, x2, y2) based on point1 and point2,
    such that x1 <= x2 and y1 <= y2.
    """
    x1 = min(point1[0], point2[0])
    x2 = max(point1[0], point2[0])
    y1 = min(point1[1], point2[1])
    y2 = max(point1[1], point2[1])
    return (x1, y1, x2, y2)


def minimum_region(point1, point2):
    """
    Returns whether the rectangle defined by the corner points point1
    and point2 exceeds the minimum dimensions.
    """
    return (
        abs(point1[0] - point2[0]) >= MIN_SELECTION_SIZE
        and abs(point1[1] - point2[1]) >= MIN_SELECTION_SIZE
    )


class Region:
    """
    Representation of a region of image that can be associated with
    a person.
    """

    def __init__(self, x1, y1, x2, y2):
        """
        Creates a new region with the specified coordinates.
        """
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2
        self.person = None
        self.mediaref = None

    def coords(self):
        """
        Returns the coordinates of the region as a 4-tuple in the
        format (x1, y1, x2, y2).
        """
        return (self.x1, self.y1, self.x2, self.y2)

    def set_coords(self, x1, y1, x2, y2):
        """
        Sets the coordinates of this region.
        """
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2

    def contains(self, x, y):
        """
        Returns whether the point with coordinates (x, y) lies insided
        this region.
        """
        return self.x1 <= x <= self.x2 and self.y1 <= y <= self.y2

    def contains_rect(self, other):
        """
        Returns whether this region fully contains the region other.
        """
        return self.contains(other.x1, other.y1) and self.contains(other.x2, other.y2)

    def area(self):
        """
        Returns the area of this region.
        """
        return abs(self.x1 - self.x2) * abs(self.y1 - self.y2)

    def intersects(self, other):
        """
        Returns whether the current region intersects other.
        """
        # assumes that x1 <= x2 and y1 <= y2
        return not (
            self.x2 < other.x1
            or self.x1 > other.x2
            or self.y2 < other.y1
            or self.y1 > other.y2
        )


class SelectionWidget(Gtk.ScrolledWindow):
    """
    A widget that displays an image and permits GIMP-like selection of regions
    within the image. The widget derives from gtk.ScrolledWindow.
    """

    __gsignals__ = {
        "region-modified": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "region-created": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "region-selected": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "selection-cleared": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "right-button-clicked": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "zoomed-in": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "zoomed-out": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self):
        """
        Creates a new selection widget.
        """
        self.multiple_selection = True

        self.loaded = False
        self.start_point_screen = None
        self.selection = None
        self.current = None
        self.in_region = None
        self.grabber = None
        self.regions = []
        self.translation = None
        self.pixbuf = None
        self.scaled_pixbuf = None
        self.scale = 1.0
        self.old_viewport_size = None

        Gtk.ScrolledWindow.__init__(self)
        self.add(self._build_gui())
        self.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)

    def _build_gui(self):
        """
        Builds and lays out the GUI of the widget.
        """
        self.image = Gtk.Image()
        self.image.set_has_tooltip(True)
        self.image.connect_after("draw", self._expose_handler)
        self.image.connect("query-tooltip", self._show_tooltip)

        self.event_box = Gtk.EventBox()
        self.event_box.connect("button-press-event", self._button_press_event)
        self.event_box.connect("button-release-event", self._button_release_event)
        self.connect("motion-notify-event", self._motion_notify_event)
        self.connect("scroll-event", self._motion_scroll_event)
        self.event_box.add_events(Gdk.EventMask.BUTTON_PRESS_MASK)
        self.event_box.add_events(Gdk.EventMask.BUTTON_RELEASE_MASK)
        self.event_box.add_events(Gdk.EventMask.POINTER_MOTION_MASK)

        self.event_box.add(self.image)

        self.viewport = Gtk.Viewport()
        self.connect("size-allocate", self._resize)
        self.viewport.add(self.event_box)
        return self.viewport

    # ======================================================
    # public field accessors
    # ======================================================

    def get_multiple_selection(self):
        """
        Return whether multiple selection is enabled.
        """
        return self.multiple_selection

    def set_multiple_selection(self, enable):
        """
        Enables or disables multiple selection.
        """
        self.multiple_selection = enable

    def is_image_loaded(self):
        """
        Returns whether an image has been loaded into this selection widget.
        """
        return self.loaded

    def set_regions(self, regions):
        """
        Sets the list of regions to be displayed in the widget.
        """
        self.regions = regions

    def get_current(self):
        """
        Returns the currently active region.
        """
        return self.current

    def set_current(self, region):
        """
        Activates the given region in the widget.
        """
        self.current = region

    def get_selection(self):
        """
        Returns the coordinates of the current selection.
        """
        return self.selection

    # ======================================================
    # loading the image
    # ======================================================

    def load_image(self, image_path):
        """
        Loads an image from a given path into this selection widget.
        """
        self.start_point_screen = None
        self.selection = None
        self.in_region = None
        self.grabber_position = None
        self.grabber_to_draw = None

        try:
            self.pixbuf = GdkPixbuf.Pixbuf.new_from_file(image_path)
            self.original_image_size = (
                self.pixbuf.get_width(),
                self.pixbuf.get_height(),
            )

            viewport_size = self.get_allocation()
            self.old_viewport_size = viewport_size
            self.scale = scale_to_fit(
                self.pixbuf.get_width(),
                self.pixbuf.get_height(),
                viewport_size.width,
                viewport_size.height,
            )
            self._rescale()
            self.loaded = True
        except (GLib.GError, OSError):
            self.show_missing()

    def show_missing(self):
        """
        Displays a 'missing image' icon in the widget.
        """
        self.pixbuf = None
        self.image.set_from_icon_name("image-missing", Gtk.IconSize.DIALOG)
        self.image.queue_draw()

    def _resize(self, *dummy):
        """
        Handles size-allocate' events from Gtk.
        """
        if self.pixbuf:
            viewport_size = self.get_allocation()
            if (
                viewport_size.height != self.old_viewport_size.height
                or viewport_size.width != self.old_viewport_size.width
                or not self.image.get_pixbuf()
            ):
                self.scale = scale_to_fit(
                    self.pixbuf.get_width(),
                    self.pixbuf.get_height(),
                    viewport_size.width,
                    viewport_size.height,
                )
                self._rescale()
                self.old_viewport_size = viewport_size
                return False

    def expander(self, *dummy):
        """Handler for expander in caller; needed because Gtk doesn't handle
        verticle expansion right
        """
        self.image.clear()
        self.image.set_size_request(2, 2)
        self.event_box.set_size_request(2, 2)
        return False

    # ======================================================
    # coordinate transformations (public methods)
    # ======================================================

    def proportional_to_real_rect(self, rect):
        """
        Translates proportional (ranging from 0 to 100) coordinates to image
        coordinates (in pixels).
        """
        x1, y1, x2, y2 = rect
        return self._proportional_to_real((x1, y1)) + self._proportional_to_real(
            (x2, y2)
        )

    def real_to_proportional_rect(self, rect):
        """
        Translates image coordinates (in pixels) to proportional (ranging
        from 0 to 100).
        """
        x1, y1, x2, y2 = rect
        return self._real_to_proportional((x1, y1)) + self._real_to_proportional(
            (x2, y2)
        )

    # ======================================================
    # widget manipulation
    # ======================================================

    def refresh(self):
        """
        Schedules a redraw of the image.
        """
        self.image.queue_draw()

    def can_zoom_in(self):
        """
        Returns whether it is possible to zoom in the image.
        """
        if self.original_image_size:
            scaled_size = (
                self.original_image_size[0] * self.scale * RESIZE_RATIO,
                self.original_image_size[1] * self.scale * RESIZE_RATIO,
            )
            return scaled_size[0] < MAX_SIZE and scaled_size[1] < MAX_SIZE
        return False

    def can_zoom_out(self):
        """
        Returns whether it is possible to zoom out the image.
        """
        if self.original_image_size:
            scaled_size = (
                self.original_image_size[0] * self.scale * RESIZE_RATIO,
                self.original_image_size[1] * self.scale * RESIZE_RATIO,
            )
            return scaled_size[0] >= MIN_SIZE and scaled_size[1] >= MIN_SIZE
        return False

    def zoom_in(self):
        """
        Zooms in the image. The zoom factor is defined by RESIZE_RATIO.
        """
        if self.can_zoom_in():
            self.scale *= RESIZE_RATIO
            self._rescale()
            self.emit("zoomed-in")

    def zoom_out(self):
        """
        Zooms out the image. The zoom factor is defined by RESIZE_RATIO.
        """
        if self.can_zoom_out():
            self.scale /= RESIZE_RATIO
            self._rescale()
            self.emit("zoomed-out")

    def select(self, region):
        """
        Highlights the given region in the image.
        """
        self.current = region
        if self.current is not None:
            self.selection = self.current.coords()
        self.image.queue_draw()

    def clear_selection(self):
        """
        Clears the selection.
        """
        self.current = None
        self.selection = None
        self.image.queue_draw()

    def find_region(self, x, y):
        """
        Given screen coordinates, find where that point is in the image.
        """
        return self._find_region(*self._screen_to_image((x, y)))

    # ======================================================
    # thumbnails
    # ======================================================

    def get_thumbnail(self, region, thumbnail_size):
        """
        Returns the thumbnail of the given region.
        """
        w = region.x2 - region.x1
        h = region.y2 - region.y1
        if w >= 1 and h >= 1 and self.pixbuf:
            subpixbuf = self.pixbuf.new_subpixbuf(region.x1, region.y1, w, h)
            size = resize_keep_aspect(w, h, *thumbnail_size)
            return subpixbuf.scale_simple(
                size[0], size[1], GdkPixbuf.InterpType.BILINEAR
            )
        else:
            return None

    # ======================================================
    # utility functions for retrieving properties
    # ======================================================

    def _get_original_image_size(self):
        """
        Returns the size of the image before scaling.
        """
        return self.original_image_size

    def _get_scaled_image_size(self):
        """
        Returns the size of images scaled with the current scaled.
        """
        unscaled_size = self._get_original_image_size()
        return (unscaled_size[0] * self.scale, unscaled_size[1] * self.scale)

    # ======================================================
    # coordinate transformations
    # ======================================================

    def _proportional_to_real(self, coord):
        """
        Translates proportional (ranging from 0 to 100) coordinates to image
        coordinates (in pixels).
        """
        w, h = self.original_image_size
        return (int(round(coord[0] * w / 100)), int(round(coord[1] * h / 100)))

    def _real_to_proportional(self, coord):
        """
        Translates image coordinates (in pixels) to proportional (ranging
        from 0 to 100).
        """
        w, h = self.original_image_size
        return (int(round(coord[0] * 100 / w)), int(round(coord[1] * 100 / h)))

    def _image_to_screen(self, coords):
        """
        Translates image coordinates to viewport coordinates using the current
        scale and viewport size.
        """
        viewport_rect = self.get_allocation()
        image_rect = self.scaled_size
        if image_rect[0] < viewport_rect.width:
            offset_x = (image_rect[0] - viewport_rect.width) / 2
        else:
            offset_x = 0.0
        if image_rect[1] < viewport_rect.height:
            offset_y = (image_rect[1] - viewport_rect.height) / 2
        else:
            offset_y = 0.0
        return (
            int(coords[0] * self.scale - offset_x),
            int(coords[1] * self.scale - offset_y),
        )

    def _screen_to_image(self, coords):
        """
        Translates viewport coordinates to original (unscaled) image coordinates
        using the current scale and viewport size.
        """
        viewport_rect = self.get_allocation()
        image_rect = self.scaled_size
        if image_rect[0] < viewport_rect.width:
            offset_x = (image_rect[0] - viewport_rect.width) / 2
        else:
            offset_x = 0.0
        if image_rect[1] < viewport_rect.height:
            offset_y = (image_rect[1] - viewport_rect.height) / 2
        else:
            offset_y = 0.0
        return (
            int((coords[0] + offset_x) / self.scale),
            int((coords[1] + offset_y) / self.scale),
        )

    def _truncate_to_image_size(self, coords):
        """
        Modifies the coordinates of the given point to ensure that it lies
        within the image. Negative values are replaced with 0, positive values
        exceeding the image dimensions - with those corresponding dimensions.
        """
        x, y = coords
        (image_width, image_height) = self._get_original_image_size()
        x = max(x, 0)
        x = min(x, image_width)
        y = max(y, 0)
        y = min(y, image_height)
        return self._proportional_to_real(self._real_to_proportional((x, y)))

    def _screen_to_truncated(self, coords):
        """
        Transforms the screen coordinates to image coordinates and truncate to
        the image size.
        """
        return self._truncate_to_image_size(self._screen_to_image(coords))

    def _rect_image_to_screen(self, rect):
        """
        Translates the coordinates of the rectangle from image to screen.
        """
        x1, y1, x2, y2 = rect
        x1, y1 = self._image_to_screen((x1, y1))
        x2, y2 = self._image_to_screen((x2, y2))
        return (x1, y1, x2, y2)

    # ======================================================
    # drawing and scaling the image
    # ======================================================

    def _expose_handler(self, widget, cr):
        """
        Handles the expose-event signal of the underlying widget.
        """
        if self.pixbuf:
            self._draw_selection(widget, cr)

    def _draw_selection(self, widget, cr):
        """
        Draws the image, the selection boxes and does the necessary
        shading.
        """
        if not self.scaled_size:
            return

        w, h = self.scaled_size
        offset_x, offset_y = self._image_to_screen((0, 0))
        offset_x -= 1
        offset_y -= 1

        if self.selection:
            x1, y1, x2, y2 = self._rect_image_to_screen(self.selection)

            # transparent shading
            self._draw_transparent_shading(cr, x1, y1, x2, y2, w, h, offset_x, offset_y)

            # selection frame
            self._draw_selection_frame(cr, x1, y1, x2, y2)

            # draw grabber
            self._draw_grabber(cr)
        else:
            # selection frame
            for region in self.regions:
                x1, y1, x2, y2 = self._rect_image_to_screen(region.coords())
                self._draw_region_frame(cr, x1, y1, x2, y2)

    def _draw_transparent_shading(self, cr, x1, y1, x2, y2, w, h, offset_x, offset_y):
        """
        Draws the shading for a selection box.
        """
        cr.set_source_rgba(1.0, 1.0, 1.0, SHADING_OPACITY)
        cr.rectangle(offset_x, offset_y, x1 - offset_x, y1 - offset_y)
        cr.rectangle(offset_x, y1, x1 - offset_x, y2 - y1)
        cr.rectangle(offset_x, y2, x1 - offset_x, h - y2 + offset_y + 1)
        cr.rectangle(x1, y2 + 1, x2 - x1 + 1, h - y2 + offset_y)
        cr.rectangle(x2 + 1, y2 + 1, w - x2 + offset_x, h - y2 + offset_y)
        cr.rectangle(x2 + 1, y1, w - x2 + offset_x, y2 - y1 + 1)
        cr.rectangle(x2 + 1, offset_y, w - x2 + offset_x, y2 - offset_y)
        cr.rectangle(x1, offset_y, x2 - x1 + 1, y1 - offset_y)
        cr.fill()

    def _draw_selection_frame(self, cr, x1, y1, x2, y2):
        """
        Draws the frame during selection.
        """
        self._draw_region_frame(cr, x1, y1, x2, y2)

    def _draw_region_frame(self, cr, x1, y1, x2, y2):
        """
        Draws a region frame.
        """
        cr.set_source_rgb(1.0, 1.0, 1.0)  # white
        cr.rectangle(x1, y1, x2 - x1, y2 - y1)
        cr.stroke()
        cr.set_source_rgb(0.0, 0.0, 1.0)  # blue
        cr.rectangle(x1 - 2, y1 - 2, x2 - x1 + 4, y2 - y1 + 4)
        cr.stroke()

    def _draw_grabber(self, cr):
        """
        Draws a grabber.
        """
        if self.selection is not None and self.grabber is not None:
            selection_rect = self._rect_image_to_screen(self.selection)
            cr.set_source_rgb(1.0, 0, 0)
            if self.grabber_position is None:
                generators = grabber_generators(selection_rect)
            elif self.grabber_position == GRABBER_INSIDE:
                generators = INNER_GRABBERS
            else:
                generators = OUTER_GRABBERS
            if self.grabber_to_draw is not None:
                generator = generators[self.grabber_to_draw]
            else:
                generator = generators[self.grabber]
            if generator is not None:
                x1, y1, x2, y2 = generator(*selection_rect)
                cr.rectangle(x1, y1, x2 - x1, y2 - y1)
            cr.stroke()

    def _rescale(self):
        """
        Recalculates the sizes using the current scale and updates
        the buffers.
        """
        self.scaled_size = (
            int(self.original_image_size[0] * self.scale),
            int(self.original_image_size[1] * self.scale),
        )
        self.scaled_image = self.pixbuf.scale_simple(
            self.scaled_size[0], self.scaled_size[1], GdkPixbuf.InterpType.BILINEAR
        )
        self.image.set_from_pixbuf(self.scaled_image)
        self.image.set_size_request(*self.scaled_size)
        self.event_box.set_size_request(*self.scaled_size)

    # ======================================================
    # managing regions
    # ======================================================

    def _find_region(self, x, y):
        """
        Finds the smallest region containing point (x, y).
        """
        result = None
        for region in self.regions:
            if region.contains(x, y):
                if result is None or result.area() > region.area():
                    result = region
        return result

    # ======================================================
    # mouse event handlers
    # ======================================================

    def _button_press_event(self, obj, event):
        """
        Handles the button-press-event signal.
        """
        if not self.is_image_loaded():
            return
        if event.button == 1:  # left button
            self.start_point_screen = (event.x, event.y)
            if (
                self.current is not None
                and self.grabber is None
                and self.multiple_selection
            ):
                self.current = None
                self.selection = None
                self.refresh()
                self.emit("selection-cleared")
        elif event.button == 3:  # right button
            # select a region, if clicked inside one
            click_point = self._screen_to_image((event.x, event.y))
            self.current = self._find_region(*click_point)
            self.selection = self.current.coords() if self.current is not None else None
            self.start_point_screen = None
            self.refresh()
            if self.current is not None:
                self.emit("region-selected")
                self.emit("right-button-clicked")
            else:
                self.emit("selection-cleared")
        return True  # don't propagate the event further

    def _button_release_event(self, obj, event):
        """
        Handles the button-release-event signal.
        """
        if not self.is_image_loaded():
            return
        if event.button == 1:
            if self.start_point_screen:
                if self.current is not None:
                    # a box is currently selected
                    if self.grabber and self.grabber != INSIDE:
                        # clicked on one of the grabbers
                        dx, dy = (
                            event.x - self.start_point_screen[0],
                            event.y - self.start_point_screen[1],
                        )
                        self.grabber_to_draw = self._modify_selection(dx, dy)
                        self.current.set_coords(*self.selection)
                        self.emit("region-modified")
                    elif self.grabber is None and self.multiple_selection:
                        # clicked outside of the grabbing area
                        self.current = None
                        self.selection = None
                        self.emit("selection-cleared")
                    else:
                        # update current selection
                        self.current.set_coords(*self.selection)
                        self.region = self.current
                        self.emit("region-modified")
                else:
                    # nothing is currently selected
                    if (
                        minimum_region(self.start_point_screen, (event.x, event.y))
                        and self._can_select()
                    ):
                        # region selection
                        region = Region(*self.selection)
                        self.regions.append(region)
                        self.current = region
                        self.emit("region-created")
                    else:
                        # nothing selected, just a click
                        click_point = self._screen_to_image(self.start_point_screen)
                        self.current = self._find_region(*click_point)
                        self.selection = (
                            self.current.coords() if self.current is not None else None
                        )
                        self.emit("region-selected")

                self.start_point_screen = None
                self.refresh()

    def _motion_notify_event(self, widget, event):
        """
        Handles the motion-notify-event signal.
        """
        if not self.is_image_loaded():
            return
        end_point_orig = self._screen_to_image((event.x, event.y))
        end_point = self._truncate_to_image_size(end_point_orig)
        if self.start_point_screen:
            # selection or dragging (mouse button pressed)
            if self.grabber is not None and self.grabber != INSIDE:
                # dragging the grabber
                dx, dy = (
                    event.x - self.start_point_screen[0],
                    event.y - self.start_point_screen[1],
                )
                self.grabber_to_draw = self._modify_selection(dx, dy)
            else:
                # making new selection
                start_point = self._screen_to_truncated(self.start_point_screen)
                self.selection = order_coordinates(start_point, end_point)

        else:
            # motion (mouse button is not pressed)
            self.in_region = self._find_region(*end_point_orig)
            if self.current is not None:
                # a box is active, so check if the pointer is inside a grabber
                rect = self._rect_image_to_screen(self.current.coords())
                self.grabber = can_grab(rect, event.x, event.y)
                if self.grabber is not None:
                    self.grabber_to_draw = self.grabber
                    self.grabber_position = grabber_position(rect)
                    self.event_box.get_window().set_cursor(CURSORS[self.grabber])
                else:
                    self.grabber_to_draw = None
                    self.grabber_position = None
                    self.event_box.get_window().set_cursor(None)
            else:
                # nothing is active
                self.grabber = None
                self.grabber_to_draw = None
                self.grabber_position = None
                self.event_box.get_window().set_cursor(None)
        self.image.queue_draw()

    def _motion_scroll_event(self, widget, event):
        """
        Handles the motion-scroll-event signal.
        """
        if not self.is_image_loaded():
            return
        if event.direction == Gdk.ScrollDirection.UP:
            self.zoom_in()
        elif event.direction == Gdk.ScrollDirection.DOWN:
            self.zoom_out()
        elif event.direction == Gdk.ScrollDirection.SMOOTH:
            if event.delta_y < 0:
                self.zoom_in()
            else:
                self.zoom_out()

    # ======================================================
    # helpers for mouse event handlers
    # ======================================================

    def _can_select(self):
        """
        Returns whether selection is currently possible, which is when
        multiple selection is enabled or otherwise when no region is
        currently selected.
        """
        return self.multiple_selection or len(self.regions) < 1

    def _modify_selection(self, dx, dy):
        """
        Changes the selection when a grabber is dragged, returns the new
        grabber if a grabber switch has happened, and the current grabber
        otherwise.
        """
        x1, y1, x2, y2 = self._rect_image_to_screen(self.current.coords())
        x1, y1, x2, y2 = MOTION_FUNCTIONS[self.grabber](x1, y1, x2, y2, dx, dy)
        (x1, y1) = self._screen_to_truncated((x1, y1))
        (x2, y2) = self._screen_to_truncated((x2, y2))
        grabber = switch_grabber(self.grabber, x1, y1, x2, y2)
        self.selection = order_coordinates((x1, y1), (x2, y2))
        return grabber

    # ======================================================
    # tooltips
    # ======================================================

    def _show_tooltip(self, widget, x, y, keyboard_mode, tooltip):
        """
        Handles the query-tooltip signal.
        """
        if self.in_region:
            person = self.in_region.person
            if person:
                name = name_displayer.display(person)
            else:
                return False
            tooltip.set_text(name)
            return True
        else:
            return False



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2006-2007  Alex Roitman
# Copyright (C) 2007-2009  Jerome Rapinat
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# plugins/quickview/Reporef.py

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------

"""
Display RepoRef for sources related to active repository
"""

from gramps.gen.simple import SimpleAccess, SimpleDoc
from gramps.gui.plug.quick import QuickTable
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


def run(database, document, repo):
    """
    Display back-references (sources) for this repository.
    """

    # setup the simple access functions

    sdb = SimpleAccess(database)
    sdoc = SimpleDoc(document)
    stab = QuickTable(sdb)

    # First we find repository and add its text

    sdoc.title("%s\n" % repo.get_name())

    # Go over all the sources that refer to this repository

    repo_handle = repo.handle
    source_list = [
        item[1] for item in database.find_backlink_handles(repo_handle, ["Source"])
    ]

    stab.columns(_("Source"), _("Type of media"), _("Call number"))
    document.has_data = False
    for source_handle in source_list:
        src = database.get_source_from_handle(source_handle)

        # Get the list of references from this source to our repo
        # (can be more than one, technically)

        for reporef in src.get_reporef_list():
            if reporef.ref == repo_handle:
                # Determine the text for this source

                media = str(reporef.get_media_type())
                call = reporef.get_call_number()

                stab.row(src.get_title(), media, call)
                document.has_data = True
    if document.has_data:
        stab.write(sdoc)
    else:
        sdoc.header1(_("Not found"))



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide autocompletion functionality.
"""

# -------------------------------------------------------------------------
#
# GNOME modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import GObject

from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext


def fill_combo(combo, data_list):
    """
    Fill a combo box with completion data
    """
    store = Gtk.ListStore(GObject.TYPE_STRING)

    for data in data_list:
        if data:
            store.append(row=[data])

    combo.set_model(store)
    combo.set_entry_text_column(0)
    completion = Gtk.EntryCompletion()
    completion.set_model(store)
    completion.set_minimum_key_length(1)
    completion.set_text_column(0)
    combo.get_child().set_completion(completion)


def fill_entry(entry, data_list):
    """
    Fill a entry with completion data
    """

    store = Gtk.ListStore(GObject.TYPE_STRING)

    for data in data_list:
        if data:
            store.append(row=[data])

    completion = Gtk.EntryCompletion()
    completion.set_model(store)
    completion.set_minimum_key_length(1)
    completion.set_text_column(0)
    entry.set_completion(completion)


# -------------------------------------------------------------------------
#
# StandardCustomSelector class
#
# -------------------------------------------------------------------------
class StandardCustomSelector:
    """
    This class provides an interface to selecting from the predefined
    options or entering custom string.

    The typical usage should be:
        type_sel = StandardCustomSelector(mapping,None,custom_key,active_key)
        whatever_table.attach(type_sel,...)
    or
        type_sel = StandardCustomSelector(mapping,cbe,custom_key,active_key)
    with the existing ComboBoxEntry cbe.

    To set up the combo box, specify the active key at creation time,
    or later (or with custom text) use:
        type_sel.set_values(i,s)

    and later, when or before the dialog is closed, do:
        (i,s) = type_sel.get_values()

    to obtain the tuple of (int,str) corresponding to the user selection.

    No selection will return (custom_key,'') if the custom key is given,
    or (None,'') if it is not given.

    The active_key determines the default selection that will be displayed
    upon widget creation. If omitted, the entry will be empty. If present,
    then no selection on the user's part will return the
    (active_key,mapping[active_key]) tuple.

    """

    def __init__(
        self,
        mapping,
        cbe=None,
        custom_key=None,
        active_key=None,
        additional=None,
        menu=None,
    ):
        """
        Constructor for the StandardCustomSelector class.

        :param cbe: Existing ComboBoxEntry widget to use.
        :type cbe: Gtk.ComboBoxEntry
        :param mapping: The mapping between integer and string constants.
        :type mapping:  dict
        :param custom_key: The key corresponding to the custom string entry
        :type custom_key:  int
        :param active_key: The key for the entry to make active upon creation
        :type active_key:  int
        """

        # set variables
        self.mapping = mapping
        self.custom_key = custom_key
        self.active_key = active_key
        self.active_index = 0
        self.additional = additional
        self.menu = menu

        # create combo box entry
        if cbe:
            assert cbe.get_has_entry()
            self.selector = cbe
        else:
            self.selector = Gtk.ComboBox(has_entry=True)

        # create models
        if menu:
            self.store = self.create_menu()
            completion_store = self.create_list()
        else:
            self.store = self.create_list()
            completion_store = self.store

        self.selector.set_model(self.store)
        self.selector.set_entry_text_column(1)

        if menu:
            for cell in self.selector.get_cells():
                self.selector.add_attribute(cell, "sensitive", 2)

        # renderer = Gtk.CellRendererText()
        # self.selector.pack_start(renderer, True)
        # self.selector.add_attribute(renderer, 'text', 1)
        # if self.active_key is not None:
        # self.selector.set_active(self.active_index)

        # make autocompletion work
        completion = Gtk.EntryCompletion()
        completion.set_model(completion_store)
        completion.set_minimum_key_length(1)
        completion.set_text_column(1)
        self.selector.get_child().set_completion(completion)

    def create_menu(self):
        """
        Create a model and fill it with a two-level tree corresponding to the
        menu.
        """
        store = Gtk.TreeStore(int, str, bool)
        for heading, items in self.menu:
            if self.active_key in items:
                parent = None
            else:
                parent = store.append(None, row=[None, _(heading), False])
            for item in items:
                store.append(parent, row=[item, self.mapping[item], True])

        if self.additional:
            parent = store.append(None, row=[None, _("Custom"), False])
            for event_type in self.additional:
                key, value = self.get_key_and_value(event_type)
                store.append(parent, row=[key, value, True])

        return store

    def create_list(self):
        """
        Create a model and fill it with a sorted flat list.
        """
        store = Gtk.ListStore(int, str)
        keys = sorted(self.mapping, key=self.by_value)
        index = 0
        for key in keys:
            if key != self.custom_key:
                store.append(row=[key, self.mapping[key]])
                if key == self.active_key:
                    self.active_index = index
                index += 1

        if self.additional:
            for event_type in self.additional:
                key, value = self.get_key_and_value(event_type)
                store.append(row=[key, value])
                if key == self.active_key:
                    self.active_index = index
                index += 1

        return store

    def by_value(self, val):
        """
        Method for sorting keys based on the values.
        """
        return glocale.sort_key(self.mapping[val])

    def get_values(self):
        """
        Get selected values.

        :returns: Returns (int,str) tuple corresponding to the selection.
        :rtype: tuple
        """
        active_iter = self.selector.get_active_iter()
        if active_iter:
            int_val = self.store.get_value(active_iter, 0)
            str_val = self.store.get_value(active_iter, 1)
            if str_val != self.mapping[int_val]:
                str_val = self.selector.get_child().get_text().strip()
        else:
            int_val = self.custom_key
            str_val = self.selector.get_child().get_text().strip()
        if str_val in iter(self.mapping.values()):
            for key in self.mapping:
                if str_val == self.mapping[key]:
                    int_val = key
                    break
        else:
            int_val = self.custom_key
        return (int_val, str_val)

    def set_values(self, val):
        """
        Set values according to given tuple.

        :param val: (int,str) tuple with the values to set.
        :type val: tuple
        """
        key, text = val
        if key in self.mapping and key != self.custom_key:
            self.store.foreach(self.set_int_value, key)
        elif self.custom_key is not None:
            self.selector.get_child().set_text(text)
        else:
            print("StandardCustomSelector.set(): Option not available:", val)

    def set_int_value(self, model, path, node, val):
        if model.get_value(node, 0) == val:
            self.selector.set_active_iter(node)
            return True
        return False

    def get_key_and_value(self, event_type):
        """
        Return the key and value for the given event type.  The event type may be
        a string representing a custom type, an (int, str) tuple or an EventType
        instance.
        """
        if isinstance(event_type, str):
            return (self.custom_key, event_type)
        elif isinstance(event_type, tuple):
            if event_type[1]:
                return (event_type[0], event_type[1])
        else:
            return (int(event_type), str(event_type))



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010  Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2011  Nick Hall
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Module providing a gramplet interface to the sidebar filters.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from typing import Any

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.plug import Gramplet
from gramps.gui.filters.sidebar import (
    PersonSidebarFilter,
    FamilySidebarFilter,
    EventSidebarFilter,
    SourceSidebarFilter,
    CitationSidebarFilter,
    PlaceSidebarFilter,
    MediaSidebarFilter,
    RepoSidebarFilter,
    NoteSidebarFilter,
)


# -------------------------------------------------------------------------
#
# Filter class
#
# -------------------------------------------------------------------------
class Filter(Gramplet):
    """
    The base class for all filter gramplets.
    """

    FILTER_CLASS: Any = None

    def init(self):
        self.filter = self.FILTER_CLASS(
            self.dbstate, self.uistate, self.__filter_clicked
        )
        self.widget = self.filter.get_widget()
        self.gui.get_container_widget().remove(self.gui.textview)
        self.gui.get_container_widget().add(self.widget)
        self.widget.show_all()

    def __filter_clicked(self):
        """
        Called when the filter apply button is clicked.
        """
        self.gui.view.generic_filter = self.filter.get_filter()
        self.gui.view.build_tree()


# -------------------------------------------------------------------------
#
# PersonFilter class
#
# -------------------------------------------------------------------------
class PersonFilter(Filter):
    """
    A gramplet providing a Person Filter.
    """

    FILTER_CLASS = PersonSidebarFilter


# -------------------------------------------------------------------------
#
# FamilyFilter class
#
# -------------------------------------------------------------------------
class FamilyFilter(Filter):
    """
    A gramplet providing a Family Filter.
    """

    FILTER_CLASS = FamilySidebarFilter


# -------------------------------------------------------------------------
#
# EventFilter class
#
# -------------------------------------------------------------------------
class EventFilter(Filter):
    """
    A gramplet providing a Event Filter.
    """

    FILTER_CLASS = EventSidebarFilter


# -------------------------------------------------------------------------
#
# SourceFilter class
#
# -------------------------------------------------------------------------
class SourceFilter(Filter):
    """
    A gramplet providing a Source Filter.
    """

    FILTER_CLASS = SourceSidebarFilter


# -------------------------------------------------------------------------
#
# CitationFilter class
#
# -------------------------------------------------------------------------
class CitationFilter(Filter):
    """
    A gramplet providing a Citation Filter.
    """

    FILTER_CLASS = CitationSidebarFilter


# -------------------------------------------------------------------------
#
# PlaceFilter class
#
# -------------------------------------------------------------------------
class PlaceFilter(Filter):
    """
    A gramplet providing a Place Filter.
    """

    FILTER_CLASS = PlaceSidebarFilter


# -------------------------------------------------------------------------
#
# MediaFilter class
#
# -------------------------------------------------------------------------
class MediaFilter(Filter):
    """
    A gramplet providing a Media Filter.
    """

    FILTER_CLASS = MediaSidebarFilter


# -------------------------------------------------------------------------
#
# RepositoryFilter class
#
# -------------------------------------------------------------------------
class RepositoryFilter(Filter):
    """
    A gramplet providing a Repository Filter.
    """

    FILTER_CLASS = RepoSidebarFilter


# -------------------------------------------------------------------------
#
# NoteFilter class
#
# -------------------------------------------------------------------------
class NoteFilter(Filter):
    """
    A gramplet providing a Note Filter.
    """

    FILTER_CLASS = NoteSidebarFilter



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2008  Stephane Charette
# Copyright (C) 2007-2008  Brian G. Matherly
# Copyright (C) 2009-2010  Gary Burton
# Contribution 2009 by     Bob Ham <rah@bash.sh>
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2011-2014  Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Family Lines, a Graphviz-based plugin for Gramps.
"""

# ------------------------------------------------------------------------
#
# python modules
#
# ------------------------------------------------------------------------
from functools import partial
import html

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".FamilyLines")

# ------------------------------------------------------------------------
#
# Gramps module
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.lib import EventRoleType, EventType, Person, PlaceType, Date
from gramps.gen.utils.file import media_path_full
from gramps.gen.utils.thumbnails import get_thumbnail_path, SIZE_NORMAL, SIZE_LARGE
from gramps.gen.plug.report import Report
from gramps.gen.plug.report import utils
from gramps.gen.plug.report import MenuReportOptions
from gramps.gen.plug.report import stdoptions
from gramps.gen.plug.menu import (
    NumberOption,
    ColorOption,
    BooleanOption,
    EnumeratedListOption,
    PersonListOption,
    SurnameColorOption,
)
from gramps.gen.utils.db import get_birth_or_fallback, get_death_or_fallback
from gramps.gen.config import config
from gramps.gen.utils.symbols import Symbols
from gramps.gen.proxy import CacheProxyDb
from gramps.gen.errors import ReportError
from gramps.gen.display.place import displayer as _pd

# ------------------------------------------------------------------------
#
# Constant options items
#
# ------------------------------------------------------------------------
_COLORS = [
    {"name": _("B&W outline"), "value": "outline"},
    {"name": _("Colored outline"), "value": "colored"},
    {"name": _("Color fill"), "value": "filled"},
]

_ARROWS = [
    {"name": _("Descendants <- Ancestors"), "value": "d"},
    {"name": _("Descendants -> Ancestors"), "value": "a"},
    {"name": _("Descendants <-> Ancestors"), "value": "da"},
    {"name": _("Descendants - Ancestors"), "value": ""},
]

_CORNERS = [
    {"name": _("None"), "value": ""},
    {"name": _("Female"), "value": "f"},
    {"name": _("Male"), "value": "m"},
    {"name": _("Both"), "value": "fm"},
]

# ------------------------------------------------------------------------
#
# A quick overview of the classes we'll be using:
#
#   class FamilyLinesOptions(MenuReportOptions)
#       - this class is created when the report dialog comes up
#       - all configuration controls for the report are created here
#
#   class FamilyLinesReport(Report)
#       - this class is created only after the user clicks on "OK"
#       - the actual report generation is done by this class
#
# ------------------------------------------------------------------------


class FamilyLinesOptions(MenuReportOptions):
    """
    Defines all of the controls necessary
    to configure the FamilyLines report.
    """

    def __init__(self, name, dbase):
        self.limit_parents = None
        self.max_parents = None
        self.limit_children = None
        self.max_children = None
        self.include_images = None
        self.image_location = None
        self.justyears = None
        self.include_dates = None
        MenuReportOptions.__init__(self, name, dbase)

    def add_menu_options(self, menu):
        # ---------------------
        category_name = _("Report Options")
        add_option = partial(menu.add_option, category_name)
        # ---------------------

        followpar = BooleanOption(
            _("Follow parents to determine " '"family lines"'), True
        )
        followpar.set_help(
            _(
                "Parents and their ancestors will be "
                'considered when determining "family lines".'
            )
        )
        add_option("followpar", followpar)

        followchild = BooleanOption(
            _("Follow children to determine " '"family lines"'), True
        )
        followchild.set_help(
            _("Children will be considered when " 'determining "family lines".')
        )
        add_option("followchild", followchild)

        remove_extra_people = BooleanOption(
            _("Try to remove extra " "people and families"), True
        )
        remove_extra_people.set_help(
            _(
                "People and families not directly "
                "related to people of interest will "
                "be removed when determining "
                '"family lines".'
            )
        )
        add_option("removeextra", remove_extra_people)

        arrow = EnumeratedListOption(_("Arrowhead direction"), "d")
        for i in range(0, len(_ARROWS)):
            arrow.add_item(_ARROWS[i]["value"], _ARROWS[i]["name"])
        arrow.set_help(_("Choose the direction that the arrows point."))
        add_option("arrow", arrow)

        color = EnumeratedListOption(_("Graph coloring"), "filled")
        for COLOR in _COLORS:
            color.add_item(COLOR["value"], COLOR["name"])
        color.set_help(
            _(
                "Males will be shown with blue, females "
                "with red, unless otherwise set above for filled. "
                "If the sex of an individual "
                "is unknown it will be shown with gray."
            )
        )
        add_option("color", color)

        roundedcorners = EnumeratedListOption(_("Rounded corners"), "")
        for i in range(0, len(_CORNERS)):
            roundedcorners.add_item(_CORNERS[i]["value"], _CORNERS[i]["name"])
        roundedcorners.set_help(
            _("Use rounded corners e.g. to differentiate " "between women and men.")
        )
        add_option("useroundedcorners", roundedcorners)

        stdoptions.add_gramps_id_option(menu, category_name, ownline=True)

        # ---------------------
        category_name = _("Report Options (2)")
        add_option = partial(menu.add_option, category_name)
        # ---------------------

        stdoptions.add_name_format_option(menu, category_name)

        stdoptions.add_private_data_option(menu, category_name, default=False)

        stdoptions.add_living_people_option(menu, category_name)

        locale_opt = stdoptions.add_localization_option(menu, category_name)

        stdoptions.add_date_format_option(menu, category_name, locale_opt)

        use_subgraphs = BooleanOption(_("Use subgraphs"), True)
        use_subgraphs.set_help(
            _(
                "Subgraphs can help Graphviz position "
                "spouses together, but with non-trivial "
                "graphs will result in longer lines and "
                "larger graphs."
            )
        )
        add_option("usesubgraphs", use_subgraphs)

        # --------------------------------
        add_option = partial(menu.add_option, _("People of Interest"))
        # --------------------------------

        person_list = PersonListOption(_("People of interest"))
        person_list.set_help(
            _(
                "People of interest are used as a starting "
                'point when determining "family lines".'
            )
        )
        add_option("gidlist", person_list)

        self.limit_parents = BooleanOption(_("Limit the number of ancestors"), False)
        self.limit_parents.set_help(_("Whether to " "limit the number of ancestors."))
        add_option("limitparents", self.limit_parents)
        self.limit_parents.connect("value-changed", self.limit_changed)

        self.max_parents = NumberOption("", 50, 10, 9999)
        self.max_parents.set_help(_("The maximum number " "of ancestors to include."))
        add_option("maxparents", self.max_parents)

        self.limit_children = BooleanOption(
            _("Limit the number " "of descendants"), False
        )
        self.limit_children.set_help(
            _("Whether to " "limit the number of descendants.")
        )
        add_option("limitchildren", self.limit_children)
        self.limit_children.connect("value-changed", self.limit_changed)

        self.max_children = NumberOption("", 50, 10, 9999)
        self.max_children.set_help(
            _("The maximum number " "of descendants to include.")
        )
        add_option("maxchildren", self.max_children)

        # --------------------
        category_name = _("Include")
        add_option = partial(menu.add_option, category_name)
        # --------------------

        self.include_dates = BooleanOption(_("Include dates"), True)
        self.include_dates.set_help(
            _("Whether to include dates for people " "and families.")
        )
        add_option("incdates", self.include_dates)
        self.include_dates.connect("value-changed", self.include_dates_changed)

        self.justyears = BooleanOption(_("Limit dates to years only"), False)
        self.justyears.set_help(
            _(
                "Prints just dates' year, neither "
                "month or day nor date approximation "
                "or interval are shown."
            )
        )
        add_option("justyears", self.justyears)

        include_places = BooleanOption(_("Include places"), True)
        include_places.set_help(
            _("Whether to include placenames for people " "and families.")
        )
        add_option("incplaces", include_places)

        include_num_children = BooleanOption(
            _("Include the number of " "children"), True
        )
        include_num_children.set_help(
            _(
                "Whether to include the number of "
                "children for families with more "
                "than 1 child."
            )
        )
        add_option("incchildcnt", include_num_children)

        self.include_images = BooleanOption(
            _("Include " "thumbnail images of people"), True
        )
        self.include_images.set_help(
            _("Whether to " "include thumbnail images of people.")
        )
        add_option("incimages", self.include_images)
        self.include_images.connect("value-changed", self.images_changed)

        self.image_location = EnumeratedListOption(_("Thumbnail location"), 0)
        self.image_location.add_item(0, _("Above the name"))
        self.image_location.add_item(1, _("Beside the name"))
        self.image_location.set_help(
            _("Where the thumbnail image " "should appear relative to the name")
        )
        add_option("imageonside", self.image_location)

        self.image_size = EnumeratedListOption(_("Thumbnail size"), SIZE_NORMAL)
        self.image_size.add_item(SIZE_NORMAL, _("Normal"))
        self.image_size.add_item(SIZE_LARGE, _("Large"))
        self.image_size.set_help(_("Size of the thumbnail image"))
        add_option("imagesize", self.image_size)

        # ----------------------------
        add_option = partial(menu.add_option, _("Family Colors"))
        # ----------------------------

        surname_color = SurnameColorOption(_("Family colors"))
        surname_color.set_help(_("Colors to use for various family lines."))
        add_option("surnamecolors", surname_color)

        # -------------------------
        add_option = partial(menu.add_option, _("Individuals"))
        # -------------------------

        color_males = ColorOption(_("Males"), "#e0e0ff")
        color_males.set_help(_("The color to use to display men."))
        add_option("colormales", color_males)

        color_females = ColorOption(_("Females"), "#ffe0e0")
        color_females.set_help(_("The color to use to display women."))
        add_option("colorfemales", color_females)

        color_other = ColorOption(_("Other"), "#94ef9e")
        color_other.set_help(
            _("The color to use to display people who are " "neither men nor women.")
        )
        add_option("colorother", color_other)

        color_unknown = ColorOption(_("Unknown"), "#e0e0e0")
        color_unknown.set_help(_("The color to use " "when the gender is unknown."))
        add_option("colorunknown", color_unknown)

        color_family = ColorOption(_("Families"), "#ffffe0")
        color_family.set_help(_("The color to use to display families."))
        add_option("colorfamilies", color_family)

        self.limit_changed()
        self.images_changed()

    def limit_changed(self):
        """
        Handle the change of limiting parents and children.
        """
        self.max_parents.set_available(self.limit_parents.get_value())
        self.max_children.set_available(self.limit_children.get_value())

    def images_changed(self):
        """
        Handle the change of including images.
        """
        self.image_location.set_available(self.include_images.get_value())
        self.image_size.set_available(self.include_images.get_value())

    def include_dates_changed(self):
        """
        Enable/disable menu items if dates are required
        """
        if self.include_dates.get_value():
            self.justyears.set_available(True)
        else:
            self.justyears.set_available(False)


# ------------------------------------------------------------------------
#
# FamilyLinesReport -- created once the user presses 'OK'
#
# ------------------------------------------------------------------------
class FamilyLinesReport(Report):
    """FamilyLines report"""

    def __init__(self, database, options, user):
        """
        Create FamilyLinesReport object that eventually produces the report.

        The arguments are:

        database     - the Gramps database instance
        options      - instance of the FamilyLinesOptions class for this report
        user         - a gen.user.User() instance
        name_format  - Preferred format to display names
        incl_private - Whether to include private data
        inc_id       - Whether to include IDs.
        living_people - How to handle living people
        years_past_death - Consider as living this many years after death
        """
        Report.__init__(self, database, options, user)

        menu = options.menu
        get_option_by_name = menu.get_option_by_name
        get_value = lambda name: get_option_by_name(name).get_value()

        self.set_locale(menu.get_option_by_name("trans").get_value())

        stdoptions.run_date_format_option(self, menu)

        stdoptions.run_private_data_option(self, menu)
        stdoptions.run_living_people_option(self, menu, self._locale)
        self.database = CacheProxyDb(self.database)
        self._db = self.database

        # initialize several convenient variables
        self._people = set()  # handle of people we need in the report
        self._families = set()  # handle of families we need in the report
        self._deleted_people = 0
        self._deleted_families = 0
        self._user = user

        self._followpar = get_value("followpar")
        self._followchild = get_value("followchild")
        self._removeextra = get_value("removeextra")
        self._gidlist = get_value("gidlist")
        self._colormales = get_value("colormales")
        self._colorfemales = get_value("colorfemales")
        self._colorother = get_value("colorother")
        self._colorunknown = get_value("colorunknown")
        self._colorfamilies = get_value("colorfamilies")
        self._limitparents = get_value("limitparents")
        self._maxparents = get_value("maxparents")
        self._limitchildren = get_value("limitchildren")
        self._maxchildren = get_value("maxchildren")
        self._incimages = get_value("incimages")
        self._imageonside = get_value("imageonside")
        self._imagesize = get_value("imagesize")
        self._useroundedcorners = get_value("useroundedcorners")
        self._usesubgraphs = get_value("usesubgraphs")
        self._incdates = get_value("incdates")
        self._just_years = get_value("justyears")
        self._incplaces = get_value("incplaces")
        self._incchildcount = get_value("incchildcnt")
        self.includeid = get_value("inc_id")

        arrow_str = get_value("arrow")
        if "d" in arrow_str:
            self._arrowheadstyle = "normal"
        else:
            self._arrowheadstyle = "none"
        if "a" in arrow_str:
            self._arrowtailstyle = "normal"
        else:
            self._arrowtailstyle = "none"

        # the gidlist is annoying for us to use since we always have to convert
        # the GIDs to either Person or to handles, so we may as well convert the
        # entire list right now and not have to deal with it ever again
        self._interest_set = set()
        if not self._gidlist:
            raise ReportError(_("Empty report"), _("You did not specify anybody"))
        for gid in self._gidlist.split():
            person = self._db.get_person_from_gramps_id(gid)
            if person:
                # option can be from another family tree, so person can be None
                self._interest_set.add(person.get_handle())

        stdoptions.run_name_format_option(self, menu)

        # convert the 'surnamecolors' string to a dictionary of names and colors
        self._surnamecolors = {}
        tmp = get_value("surnamecolors")
        if tmp.find("\xb0") >= 0:
            # new style delimiter (see bug report #2162)
            tmp = tmp.split("\xb0")
        else:
            # old style delimiter
            tmp = tmp.split(" ")

        while len(tmp) > 1:
            surname = tmp.pop(0).encode("iso-8859-1", "xmlcharrefreplace")
            colour = tmp.pop(0)
            self._surnamecolors[surname] = colour

        self._colorize = get_value("color")

    def begin_report(self):
        """
        Inherited method; called by report() in _ReportDialog.py

        This is where we'll do all of the work of figuring out who
        from the database is going to be output into the report
        """

        # starting with the people of interest, we then add parents:
        self._people.clear()
        self._families.clear()
        if self._followpar:
            self.find_parents()

            if self._removeextra:
                self.remove_uninteresting_parents()

        # ...and/or with the people of interest we add their children:
        if self._followchild:
            self.find_children()
        # once we get here we have a full list of people
        # and families that we need to generate a report

    def write_report(self):
        """
        Inherited method; called by report() in _ReportDialog.py
        """

        # now that begin_report() has done the work, output what we've
        # obtained into whatever file or format the user expects to use

        self.doc.add_comment(
            "# %s %d"
            % (self._("Number of people in database:"), self._db.get_number_of_people())
        )
        self.doc.add_comment(
            "# %s %d" % (self._("Number of people of interest:"), len(self._people))
        )
        self.doc.add_comment(
            "# %s %d"
            % (
                self._("Number of families in database:"),
                self._db.get_number_of_families(),
            )
        )
        self.doc.add_comment(
            "# %s %d" % (self._("Number of families of interest:"), len(self._families))
        )
        if self._removeextra:
            self.doc.add_comment(
                "# %s %d" % (self._("Additional people removed:"), self._deleted_people)
            )
            self.doc.add_comment(
                "# %s %d"
                % (self._("Additional families removed:"), self._deleted_families)
            )
        self.doc.add_comment("# %s" % self._("Initial list of people of interest:"))
        for handle in self._interest_set:
            person = self._db.get_person_from_handle(handle)
            gid = person.get_gramps_id()
            name = person.get_primary_name().get_regular_name()
            # Translators: needed for Arabic, ignore otherwise
            id_n = self._("%(str1)s, %(str2)s") % {"str1": gid, "str2": name}
            self.doc.add_comment("# -> " + id_n)

        self.write_people()
        self.write_families()

    def find_parents(self):
        """find the parents"""
        # we need to start with all of our "people of interest"
        ancestors_not_yet_processed = set(self._interest_set)

        # now we find all the immediate ancestors of our people of interest

        while ancestors_not_yet_processed:
            handle = ancestors_not_yet_processed.pop()

            # One of 2 things can happen here:
            #   1) we already know about this person and he/she is already
            #      in our list
            #   2) this is someone new, and we need to remember him/her
            #
            # In the first case, there isn't anything else to do, so we simply
            # go back to the top and pop the next person off the list.
            #
            # In the second case, we need to add this person to our list, and
            # then go through all of the parents this person has to find more
            # people of interest.

            if handle not in self._people:
                person = self._db.get_person_from_handle(handle)

                # remember this person!
                self._people.add(handle)

                # see if a family exists between this person and someone else
                # we have on our list of people we're going to output -- if
                # there is a family, then remember it for when it comes time
                # to link spouses together
                for family_handle in person.get_family_handle_list():
                    family = self._db.get_family_from_handle(family_handle)
                    if not family:
                        continue
                    spouse_handle = utils.find_spouse(person, family)
                    if spouse_handle:
                        if (
                            spouse_handle in self._people
                            or spouse_handle in ancestors_not_yet_processed
                        ):
                            self._families.add(family_handle)

                # if we have a limit on the number of people, and we've
                # reached that limit, then don't attempt to find any
                # more ancestors
                if self._limitparents and (
                    self._maxparents
                    < len(ancestors_not_yet_processed) + len(self._people)
                ):
                    # get back to the top of the while loop so we can finish
                    # processing the people queued up in the "not yet
                    # processed" list
                    continue

                # queue the parents of the person we're processing
                for family_handle in person.get_parent_family_handle_list():
                    family = self._db.get_family_from_handle(family_handle)

                    father_handle = family.get_father_handle()
                    if father_handle:
                        father = self._db.get_person_from_handle(father_handle)
                        if father:
                            ancestors_not_yet_processed.add(father_handle)
                            self._families.add(family_handle)

                    mother_handle = family.get_mother_handle()
                    if mother_handle:
                        mother = self._db.get_person_from_handle(mother_handle)
                        if mother:
                            ancestors_not_yet_processed.add(mother_handle)
                            self._families.add(family_handle)

    def remove_uninteresting_parents(self):
        """remove any uninteresting parents"""
        # start with all the people we've already identified
        unprocessed_parents = set(self._people)

        while len(unprocessed_parents) > 0:
            handle = unprocessed_parents.pop()
            person = self._db.get_person_from_handle(handle)
            if not person:
                continue

            # There are a few things we're going to need,
            # so look it all up right now; such as:
            # - who is the child?
            # - how many children?
            # - parents?
            # - spouse?
            # - is a person of interest?
            # - spouse of a person of interest?
            # - same surname as a person of interest?
            # - spouse has the same surname as a person of interest?

            child_handle = None
            child_count = 0
            spouse_handle = None
            spouse_count = 0
            father_handle = None
            mother_handle = None
            spouse_father_handle = None
            spouse_mother_handle = None
            spouse_surname = ""
            surname = person.get_primary_name().get_surname()
            surname = surname.encode("iso-8859-1", "xmlcharrefreplace")

            # first we get the person's father and mother
            for family_handle in person.get_parent_family_handle_list():
                family = self._db.get_family_from_handle(family_handle)
                handle = family.get_father_handle()
                if handle in self._people:
                    father_handle = handle
                handle = family.get_mother_handle()
                if handle in self._people:
                    mother_handle = handle

            # now see how many spouses this person has
            for family_handle in person.get_family_handle_list():
                family = self._db.get_family_from_handle(family_handle)
                handle = utils.find_spouse(person, family)
                if handle in self._people:
                    spouse_count += 1
                    spouse = self._db.get_person_from_handle(handle)
                    spouse_handle = handle
                    spouse_surname = spouse.get_primary_name().get_surname()
                    spouse_surname = spouse_surname.encode(
                        "iso-8859-1", "xmlcharrefreplace"
                    )

                    # see if the spouse has parents
                    if not spouse_father_handle and not spouse_mother_handle:
                        for family_handle in spouse.get_parent_family_handle_list():
                            family = self._db.get_family_from_handle(family_handle)
                            handle = family.get_father_handle()
                            if handle in self._people:
                                spouse_father_handle = handle
                            handle = family.get_mother_handle()
                            if handle in self._people:
                                spouse_mother_handle = handle

            # get the number of children that we think might be interesting
            for family_handle in person.get_family_handle_list():
                family = self._db.get_family_from_handle(family_handle)
                for child_ref in family.get_child_ref_list():
                    if child_ref.ref in self._people:
                        child_count += 1
                        child_handle = child_ref.ref

            # we now have everything we need -- start looking for reasons
            # why this is a person we need to keep in our list, and loop
            # back to the top as soon as a reason is discovered

            # if this person has many children of interest, then we
            # automatically keep this person
            if child_count > 1:
                continue

            # if this person has many spouses of interest, then we
            # automatically keep this person
            if spouse_count > 1:
                continue

            # if this person has parents, then we automatically keep
            # this person
            if father_handle is not None or mother_handle is not None:
                continue

            # if the spouse has parents, then we automatically keep
            # this person
            if spouse_father_handle is not None or spouse_mother_handle is not None:
                continue

            # if this is a person of interest, then we automatically keep
            if person.get_handle() in self._interest_set:
                continue

            # if the spouse is a person of interest, then we keep
            if spouse_handle in self._interest_set:
                continue

            # if the surname (or the spouse's surname) matches a person
            # of interest, then we automatically keep this person
            keep_this_person = False
            for person_of_interest_handle in self._interest_set:
                person_of_interest = self._db.get_person_from_handle(
                    person_of_interest_handle
                )
                surname_of_interest = person_of_interest.get_primary_name()
                surname_of_interest = surname_of_interest.get_surname().encode(
                    "iso-8859-1", "xmlcharrefreplace"
                )
                if (
                    surname_of_interest == surname
                    or surname_of_interest == spouse_surname
                ):
                    keep_this_person = True
                    break

            if keep_this_person:
                continue

            # if we have a special colour to use for this person,
            # then we automatically keep this person
            if surname in self._surnamecolors:
                continue

            # if we have a special colour to use for the spouse,
            # then we automatically keep this person
            if spouse_surname in self._surnamecolors:
                continue

            # took us a while,
            # but if we get here then we can remove this person
            self._deleted_people += 1
            self._people.remove(person.get_handle())

            # we can also remove any families to which this person belonged
            for family_handle in person.get_family_handle_list():
                if family_handle in self._families:
                    self._deleted_families += 1
                    self._families.remove(family_handle)

            # if we have a spouse, then ensure we queue up the spouse
            if spouse_handle:
                if spouse_handle not in unprocessed_parents:
                    unprocessed_parents.add(spouse_handle)

            # if we have a child, then ensure we queue up the child
            if child_handle:
                if child_handle not in unprocessed_parents:
                    unprocessed_parents.add(child_handle)

    def find_children(self):
        """find any children"""
        # we need to start with all of our "people of interest"
        children_not_yet_processed = set(self._interest_set)
        children_to_include = set()

        # now we find all the children of our people of interest

        while len(children_not_yet_processed) > 0:
            handle = children_not_yet_processed.pop()

            if handle not in children_to_include:
                person = self._db.get_person_from_handle(handle)

                # remember this person!
                children_to_include.add(handle)

                # if we have a limit on the number of people, and we've
                # reached that limit, then don't attempt to find any
                # more children
                if self._limitchildren and (
                    self._maxchildren
                    < len(children_not_yet_processed) + len(children_to_include)
                ):
                    # get back to the top of the while loop
                    # so we can finish processing the people
                    # queued up in the "not yet processed" list
                    continue

                # iterate through this person's families
                for family_handle in person.get_family_handle_list():
                    family = self._db.get_family_from_handle(family_handle)

                    # queue up any children from this person's family
                    for childref in family.get_child_ref_list():
                        child = self._db.get_person_from_handle(childref.ref)
                        children_not_yet_processed.add(child.get_handle())
                        self._families.add(family_handle)

                    # include the spouse from this person's family
                    spouse_handle = utils.find_spouse(person, family)
                    if spouse_handle:
                        children_to_include.add(spouse_handle)
                        self._families.add(family_handle)

        # we now merge our temp set "children_to_include" into our master set
        self._people.update(children_to_include)

    def write_people(self):
        """write the people"""

        self.doc.add_comment("")

        # If we're going to attempt to include images, then use the HTML style
        # of .gv file.
        use_html_output = False
        if self._incimages:
            use_html_output = True

        # get birth and death symbols
        symbols = Symbols()
        death_idx = config.get("utf8.death-symbol")
        death_symbol = symbols.get_death_symbol_for_char(death_idx)
        birth_symbol = symbols.get_symbol_for_string(symbols.SYMBOL_BIRTH)

        # loop through all the people we need to output
        for handle in sorted(self._people):  # enable a diff
            person = self._db.get_person_from_handle(handle)
            name = self._name_display.display(person)
            p_id = person.get_gramps_id()

            # figure out what colour to use
            gender = person.get_gender()
            colour = self._colorunknown
            if gender == Person.MALE:
                colour = self._colormales
            elif gender == Person.FEMALE:
                colour = self._colorfemales
            elif gender == Person.OTHER:
                colour = self._colorother

            # see if we have surname colours that match this person
            surname = person.get_primary_name().get_surname()
            surname = surname.encode("iso-8859-1", "xmlcharrefreplace")
            if surname in self._surnamecolors:
                colour = self._surnamecolors[surname]

            # see if we have a birth/death or fallback dates we can use
            if self._incdates or self._incplaces:
                bth_event = get_birth_or_fallback(self._db, person)
                dth_event = get_death_or_fallback(self._db, person)
            else:
                bth_event = None
                dth_event = None

            # output the birth or fallback event
            birth_str = None
            if bth_event and self._incdates:
                date = bth_event.get_date_object()
                if self._just_years and date.get_year_valid():
                    birth_str = self.get_date(Date(date.get_year()))  # localized year
                else:
                    birth_str = self.get_date(date)

            # get birth place (one of:  hamlet, village, town, city, parish,
            # county, province, region, state or country)
            birthplace = None
            if bth_event and self._incplaces:
                birthplace = self.get_event_place(bth_event)

            # see if we have a deceased date we can use
            death_str = None
            if dth_event and self._incdates:
                date = dth_event.get_date_object()
                if self._just_years and date.get_year_valid():
                    death_str = self.get_date(Date(date.get_year()))  # localized year
                else:
                    death_str = self.get_date(date)

            # get death place (one of:  hamlet, village, town, city, parish,
            # county, province, region, state or country)
            deathplace = None
            if dth_event and self._incplaces:
                deathplace = self.get_event_place(dth_event)

            # see if we have an image to use for this person
            image_path = None
            if self._incimages:
                media_list = person.get_media_list()
                if len(media_list) > 0:
                    media_handle = media_list[0].get_reference_handle()
                    media = self._db.get_media_from_handle(media_handle)
                    media_mime_type = media.get_mime_type()
                    if media_mime_type[0:5] == "image":
                        image_path = get_thumbnail_path(
                            media_path_full(self._db, media.get_path()),
                            rectangle=media_list[0].get_rectangle(),
                            size=self._imagesize,
                        )

            # put the label together and output this person
            label = ""
            line_delimiter = "\\n"
            if use_html_output:
                line_delimiter = "<BR/>"

            # if we have an image, then start an HTML table;
            # remember to close the table afterwards!
            if image_path:
                label = (
                    '<TABLE BORDER="0" CELLSPACING="2" CELLPADDING="0" '
                    'CELLBORDER="0"><TR><TD><IMG SRC="%s"/></TD>' % image_path
                )
                if self._imageonside == 0:
                    label += "</TR><TR>"
                label += "<TD>"

            # at the very least, the label must have the person's name
            label += html.escape(name)
            if self.includeid == 1:  # same line
                label += " (%s)" % p_id
            elif self.includeid == 2:  # own line
                label += "%s(%s)" % (line_delimiter, p_id)

            if birth_str or birthplace:
                label += "%s%s " % (line_delimiter, birth_symbol)
                if birth_str:
                    label += "%s" % birth_str
                if birth_str and birthplace:
                    label += " - "
                if birthplace:
                    label += "%s" % birthplace
            if death_str or deathplace:
                label += "%s%s " % (line_delimiter, death_symbol)
                if death_str:
                    label += "%s" % death_str
                if death_str and deathplace:
                    label += " - "
                if deathplace:
                    label += "%s" % deathplace

            # see if we have a table that needs to be terminated
            if image_path:
                label += "</TD></TR></TABLE>"
            else:
                # non html label is enclosed by "" so escape other "
                label = label.replace('"', '\\"')

            shape = "box"
            style = "solid"
            border = colour
            fill = colour

            # do not use colour if this is B&W outline
            if self._colorize == "outline":
                border = ""
                fill = ""

            if gender == person.FEMALE and ("f" in self._useroundedcorners):
                style = "rounded"
            elif gender == person.MALE and ("m" in self._useroundedcorners):
                style = "rounded"
            elif gender == person.UNKNOWN:
                shape = "hexagon"

            # if we're filling the entire node:
            if self._colorize == "filled":
                style += ",filled"
                border = ""

            # we're done -- add the node
            self.doc.add_node(
                p_id,
                label=label,
                shape=shape,
                color=border,
                style=style,
                fillcolor=fill,
                htmloutput=use_html_output,
            )

    def write_families(self):
        """write the families"""

        self.doc.add_comment("")
        ngettext = self._locale.translation.ngettext  # to see "nearby" comments

        # loop through all the families we need to output
        for family_handle in sorted(self._families):  # enable a diff
            family = self._db.get_family_from_handle(family_handle)
            fgid = family.get_gramps_id()

            # figure out a wedding date or placename we can use
            wedding_date = None
            wedding_place = None
            if self._incdates or self._incplaces:
                for event_ref in family.get_event_ref_list():
                    event = self._db.get_event_from_handle(event_ref.ref)
                    if event.get_type() == EventType.MARRIAGE and (
                        event_ref.get_role() == EventRoleType.FAMILY
                        or event_ref.get_role() == EventRoleType.PRIMARY
                    ):
                        # get the wedding date
                        if self._incdates:
                            date = event.get_date_object()
                            if self._just_years and date.get_year_valid():
                                wedding_date = self.get_date(  # localized year
                                    Date(date.get_year())
                                )
                            else:
                                wedding_date = self.get_date(date)
                        # get the wedding location
                        if self._incplaces:
                            wedding_place = self.get_event_place(event)
                        break

            # figure out the number of children (if any)
            children_str = None
            if self._incchildcount:
                child_count = len(family.get_child_ref_list())
                if child_count >= 1:
                    # Translators: leave all/any {...} untranslated
                    children_str = ngettext(
                        "{number_of} child", "{number_of} children", child_count
                    ).format(number_of=child_count)

            label = ""
            fgid_already = False
            if wedding_date:
                if label != "":
                    label += "\\n"
                label += "%s" % wedding_date
                if self.includeid == 1 and not fgid_already:  # same line
                    label += " (%s)" % fgid
                    fgid_already = True
            if wedding_place:
                if label != "":
                    label += "\\n"
                label += "%s" % wedding_place
                if self.includeid == 1 and not fgid_already:  # same line
                    label += " (%s)" % fgid
                    fgid_already = True
            if self.includeid == 1 and not label:
                label = "(%s)" % fgid
                fgid_already = True
            elif self.includeid == 2 and not label:  # own line
                label = "(%s)" % fgid
                fgid_already = True
            elif self.includeid == 2 and label and not fgid_already:
                label += "\\n(%s)" % fgid
                fgid_already = True
            if children_str:
                if label != "":
                    label += "\\n"
                label += "%s" % children_str
                if self.includeid == 1 and not fgid_already:  # same line
                    label += " (%s)" % fgid
                    fgid_already = True

            shape = "ellipse"
            style = "solid"
            border = self._colorfamilies
            fill = self._colorfamilies

            # do not use colour if this is B&W outline
            if self._colorize == "outline":
                border = ""
                fill = ""

            # if we're filling the entire node:
            if self._colorize == "filled":
                style += ",filled"
                border = ""

            # we're done -- add the node
            self.doc.add_node(fgid, label, shape, border, style, fill)

        # now that we have the families written,
        # go ahead and link the parents and children to the families
        for family_handle in self._families:
            # get the parents for this family
            family = self._db.get_family_from_handle(family_handle)
            fgid = family.get_gramps_id()
            father_handle = family.get_father_handle()
            mother_handle = family.get_mother_handle()

            self.doc.add_comment("")

            if self._usesubgraphs and father_handle and mother_handle:
                self.doc.start_subgraph(fgid)

            # see if we have a father to link to this family
            if father_handle:
                if father_handle in self._people:
                    father = self._db.get_person_from_handle(father_handle)
                    father_rn = father.get_primary_name().get_regular_name()
                    comment = self._("father: %s") % father_rn
                    self.doc.add_link(
                        father.get_gramps_id(),
                        fgid,
                        "",
                        self._arrowheadstyle,
                        self._arrowtailstyle,
                        comment=comment,
                    )

            # see if we have a mother to link to this family
            if mother_handle:
                if mother_handle in self._people:
                    mother = self._db.get_person_from_handle(mother_handle)
                    mother_rn = mother.get_primary_name().get_regular_name()
                    comment = self._("mother: %s") % mother_rn
                    self.doc.add_link(
                        mother.get_gramps_id(),
                        fgid,
                        "",
                        self._arrowheadstyle,
                        self._arrowtailstyle,
                        comment=comment,
                    )

            if self._usesubgraphs and father_handle and mother_handle:
                self.doc.end_subgraph()

            # link the children to the family
            for childref in family.get_child_ref_list():
                if childref.ref in self._people:
                    child = self._db.get_person_from_handle(childref.ref)
                    child_rn = child.get_primary_name().get_regular_name()
                    comment = self._("child: %s") % child_rn
                    self.doc.add_link(
                        fgid,
                        child.get_gramps_id(),
                        "",
                        self._arrowheadstyle,
                        self._arrowtailstyle,
                        comment=comment,
                    )

    def get_event_place(self, event):
        """get the place of the event"""
        place_text = ""
        place_handle = event.get_place_handle()
        if place_handle:
            place = self._db.get_place_from_handle(place_handle)
            if place:
                place_text = _pd.display(self._db, place)
                place_text = html.escape(place_text)
        return place_text

    def get_date(self, date):
        """return a formatted date"""
        return html.escape(self._get_date(date))



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Gary Burton
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Tools/Utilities/Relationship Calculator"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# GNOME libraries
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.display.name import displayer as name_displayer
from gramps.gui.managedwindow import ManagedWindow
from gramps.gui.views.treemodels import PeopleBaseModel, PersonTreeModel
from gramps.plugins.lib.libpersonview import BasePersonView
from gramps.gen.relationship import get_relationship_calculator
from gramps.gen.const import URL_MANUAL_PAGE
from gramps.gui.display import display_help

from gramps.gui.dialog import ErrorDialog
from gramps.gui.plug import tool
from gramps.gui.glade import Glade

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

column_names = [column[0] for column in BasePersonView.COLUMNS]
WIKI_HELP_PAGE = URL_MANUAL_PAGE + "_-_Tools"
WIKI_HELP_SEC = _("Relationship Calculator")


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
class RelCalc(tool.Tool, ManagedWindow):
    def __init__(self, dbstate, user, options_class, name, callback=None):
        uistate = user.uistate
        """
        Relationship calculator class.
        """

        tool.Tool.__init__(self, dbstate, options_class, name)
        ManagedWindow.__init__(self, uistate, [], self.__class__)

        # set the columns to see
        for data in BasePersonView.CONFIGSETTINGS:
            if data[0] == "columns.rank":
                colord = data[1]
            elif data[0] == "columns.visible":
                colvis = data[1]
            elif data[0] == "columns.size":
                colsize = data[1]
        self.colord = []
        for col, size in zip(colord, colsize):
            if col in colvis:
                self.colord.append((1, col, size))
            else:
                self.colord.append((0, col, size))

        self.dbstate = dbstate
        self.relationship = get_relationship_calculator(glocale)
        self.relationship.connect_db_signals(dbstate)

        self.glade = Glade()
        self.person = self.db.get_person_from_handle(uistate.get_active("Person"))
        name = ""
        if self.person:
            name = name_displayer.display(self.person)
        self.title = _("Relationship calculator: %(person_name)s") % {
            "person_name": name
        }
        window = self.glade.toplevel
        self.titlelabel = self.glade.get_object("title")
        self.set_window(
            window,
            self.titlelabel,
            _("Relationship to %(person_name)s") % {"person_name": name},
            self.title,
        )
        self.setup_configs("interface.relcalc", 600, 400)

        self.tree = self.glade.get_object("peopleList")
        self.text = self.glade.get_object("text1")
        self.textbuffer = Gtk.TextBuffer()
        self.text.set_buffer(self.textbuffer)

        self.model = PersonTreeModel(self.db, uistate)
        self.tree.set_model(self.model)

        self.tree.connect("key-press-event", self._key_press)
        self.selection = self.tree.get_selection()
        self.selection.set_mode(Gtk.SelectionMode.SINGLE)

        # keep reference of column so garbage collection works
        self.columns = []
        for pair in self.colord:
            if not pair[0]:
                continue
            name = column_names[pair[1]]
            column = Gtk.TreeViewColumn(name, Gtk.CellRendererText(), markup=pair[1])
            column.set_resizable(True)
            column.set_min_width(60)
            column.set_sizing(Gtk.TreeViewColumnSizing.GROW_ONLY)
            self.tree.append_column(column)
            # keep reference of column so garbage collection works
            self.columns.append(column)

        self.sel = self.tree.get_selection()
        self.changedkey = self.sel.connect("changed", self.on_apply_clicked)
        self.closebtn = self.glade.get_object("button5")
        self.closebtn.connect("clicked", self.close)
        help_btn = self.glade.get_object("help_btn")
        help_btn.connect(
            "clicked", lambda x: display_help(WIKI_HELP_PAGE, WIKI_HELP_SEC)
        )

        if not self.person:
            self.window.hide()
            ErrorDialog(
                _("Active person has not been set"),
                _(
                    "You must select an active person for this "
                    "tool to work properly."
                ),
                parent=uistate.window,
            )
            self.close()
            return

        self.show()

    def close(self, *obj):
        """Close relcalc tool. Remove non-gtk connections so garbage
        collection can do its magic.
        """
        self.relationship.disconnect_db_signals(self.dbstate)
        self.sel.disconnect(self.changedkey)
        ManagedWindow.close(self, *obj)

    def build_menu_names(self, obj):
        return (_("Relationship Calculator tool"), None)

    def on_apply_clicked(self, obj):
        model, iter_ = self.tree.get_selection().get_selected()
        if not iter_:
            return

        other_person = None
        handle = model.get_handle_from_iter(iter_)
        if handle:
            other_person = self.db.get_person_from_handle(handle)
        if other_person is None:
            self.textbuffer.set_text("")
            return

        # now determine the relation, and print it out
        rel_strings, common_an = self.relationship.get_all_relationships(
            self.db, self.person, other_person
        )

        p1 = name_displayer.display(self.person)
        p2 = name_displayer.display(other_person)

        text = []
        if other_person is None:
            pass
        elif self.person.handle == other_person.handle:
            rstr = _("%(person)s and %(active_person)s are the same person.") % {
                "person": p1,
                "active_person": p2,
            }
            text.append((rstr, ""))
        elif len(rel_strings) == 0:
            rstr = _("%(person)s and %(active_person)s are not related.") % {
                "person": p2,
                "active_person": p1,
            }
            text.append((rstr, ""))

        for rel_string, common in zip(rel_strings, common_an):
            rstr = _("%(person)s is the %(relationship)s of %(active_person)s.") % {
                "person": p2,
                "relationship": rel_string,
                "active_person": p1,
            }
            length = len(common)
            if length == 1:
                person = self.db.get_person_from_handle(common[0])
                if common[0] in [other_person.handle, self.person.handle]:
                    commontext = ""
                else:
                    name = name_displayer.display(person)
                    commontext = " " + _("Their common ancestor is %s.") % name
            elif length == 2:
                p1c = self.db.get_person_from_handle(common[0])
                p2c = self.db.get_person_from_handle(common[1])
                p1str = name_displayer.display(p1c)
                p2str = name_displayer.display(p2c)
                commontext = " " + _(
                    "Their common ancestors are %(ancestor1)s and %(ancestor2)s."
                ) % {"ancestor1": p1str, "ancestor2": p2str}
            elif length > 2:
                index = 0
                commontext = " " + _("Their common ancestors are: ")
                for person_handle in common:
                    person = self.db.get_person_from_handle(person_handle)
                    if index:
                        # TODO for Arabic, should the next comma be translated?
                        commontext += ", "
                    commontext += name_displayer.display(person)
                    index += 1
                commontext += "."
            else:
                commontext = ""
            text.append((rstr, commontext))

        textval = ""
        for val in text:
            textval += "%s %s\n" % (val[0], val[1])
        self.textbuffer.set_text(textval)

    def _key_press(self, obj, event):
        if event.keyval in (Gdk.KEY_Return, Gdk.KEY_KP_Enter):
            store, paths = self.selection.get_selected_rows()
            if paths and len(paths[0]) == 1:
                if self.tree.row_expanded(paths[0]):
                    self.tree.collapse_row(paths[0])
                else:
                    self.tree.expand_row(paths[0], 0)
                return True
        return False


# ------------------------------------------------------------------------
#
#
#
# ------------------------------------------------------------------------
class RelCalcOptions(tool.ToolOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, person_id=None):
        tool.ToolOptions.__init__(self, name, person_id)



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2005  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Andrew I Baznikin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Written by Alex Roitman, largely based on relationship.py by Don Allingham
# and on valuable input from Frode Jemtland
# Tilraun til aðlögunar: Sveinn í Felli, 02/2016
"""
Icelandic-Specific classes for relationships.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

from gramps.gen.lib import Person
import gramps.gen.relationship

# -------------------------------------------------------------------------

_cousin_level = [
    "",
    "",  # brother/sister, frændi/frænka -- these are taken care of separately
    "þremenningur",
    "fjórmenningur",
    "fimmmenningur",
    "sexmenningur",
    "sjömenningur",
    "áttmenningur",
    "nímenningur",
    "tímenningur",
    "ellefumenningur",
    "tólfmenningur",
    "þrettánmenningur",
    "fjórtánmenningur",
    "fimtánmenningur",
    "sekstenmenning",
    "syttenmenning",
    "attenmenning",
    "nittenmenning",
    "tyvemenning",
]

_cousin_terms = _cousin_level + ["frændi", "frænka"]


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
class RelationshipCalculator(gramps.gen.relationship.RelationshipCalculator):
    """
    RelationshipCalculator Class
    """

    def __init__(self):
        gramps.gen.relationship.RelationshipCalculator.__init__(self)

    def get_parents(self, level):
        if level == 0:
            return "foreldri"
        else:
            return "forfaðir í %d-tu kynslóð" % (level + 1)

    def get_cousin(self, level):
        if level > len(_cousin_level) - 1:
            # FIXME: use Norwegian term term here,
            # UPDATED by Frode: unsure about the expretion "en fjern slektning", should it be just "fjern slektning".
            # Need to see it used in the program to get the understanding.
            return "fjarskyldur ættingi"
        else:
            return _cousin_level[level]

    def pair_up(self, rel_list):
        result = []
        item = ""
        for word in rel_list[:]:
            if not word:
                continue
            if word in _cousin_terms:
                if item:
                    result.append(item)
                    item = ""
                result.append(word)
                continue
            if item:
                if word == "synir":
                    word = "sonur"
                result.append(item + word)
                item = ""
            else:
                item = word
        if item:
            result.append(item)
        gen_result = [item + "s" for item in result[0:-1]]
        return " ".join(gen_result + result[-1:])

    def get_direct_ancestor(self, person, rel_string):
        result = []
        for rel in rel_string:
            if rel == "f":
                result.append("faðir")
            else:
                result.append("móðir")
        return self.pair_up(result)

    def get_direct_descendant(self, person, rel_string):
        result = []
        for ix in range(len(rel_string) - 2, -1, -1):
            if rel_string[ix] == "f":
                result.append("synir")
            else:
                result.append("dætur")
        if person == Person.MALE:
            result.append("sonur")
        else:
            result.append("dóttir")
        return self.pair_up(result)

    def get_ancestors_cousin(self, person, rel_string_long, rel_string_short):
        result = []
        removed = len(rel_string_long) - len(rel_string_short)
        level = len(rel_string_short) - 1
        for ix in range(removed):
            if rel_string_long[ix] == "f":
                result.append("faðir")
            else:
                result.append("móðir")
        if level > 1:
            result.append(self.get_cousin(level))
        elif person == Person.MALE:
            result.append("frændi")
        else:
            result.append("frænka")
        main_string = self.get_two_way_rel(person, rel_string_short, rel_string_long)
        aux_string = self.pair_up(result)
        return "%s (%s)" % (main_string, aux_string)

    def get_cousins_descendant(self, person, rel_string_long, rel_string_short):
        result = []
        aux_string = ""
        removed = len(rel_string_long) - len(rel_string_short) - 1
        level = len(rel_string_short) - 1
        if level > 1:  # Cousin terms without gender
            result.append(self.get_cousin(level))
        elif level == 1:  # gender-dependent fetter/kusine
            if rel_string_long[removed] == "f":
                result.append("fændi")
            else:
                result.append("frænka")
        elif rel_string_long[removed] == "f":
            result.append("bróðir")
        else:
            result.append("systir")
        for ix in range(removed - 1, -1, -1):
            if rel_string_long[ix] == "f":
                result.append("sonur")
            else:
                result.append("dóttir")
        if person == Person.MALE:
            result.append("sonur")
        else:
            result.append("dóttir")
        main_string = self.get_two_way_rel(person, rel_string_long, rel_string_short)
        if level:
            aux_string = " (%s)" % self.pair_up(result)
        return "%s%s" % (main_string, aux_string)

    def get_ancestors_brother(self, rel_string):
        result = []
        for ix in range(len(rel_string) - 1):
            if rel_string[ix] == "f":
                result.append("föður")
            else:
                result.append("móður")
        result.append("bróðir")
        return self.pair_up(result)

    def get_ancestors_sister(self, rel_string):
        result = []
        for ix in range(len(rel_string) - 1):
            if rel_string[ix] == "f":
                result.append("föður")
            else:
                result.append("móður")
        result.append("systir")
        return self.pair_up(result)

    def get_two_way_rel(self, person, first_rel_string, second_rel_string):
        result = []
        for ix in range(len(second_rel_string) - 1):
            if second_rel_string[ix] == "f":
                result.append("föður")
            else:
                result.append("móður")
        if len(first_rel_string) > 1:
            if first_rel_string[-2] == "f":
                result.append("bróðir")
            else:
                result.append("systir")
            for ix in range(len(first_rel_string) - 3, -1, -1):
                if first_rel_string[ix] == "f":
                    result.append("sønne")
                else:
                    result.append("datter")
            if person == Person.MALE:
                result.append("sonur")
            else:
                result.append("dóttir")
        else:
            if person == Person.MALE:
                result.append("bróðir")
            else:
                result.append("systir")
        return self.pair_up(result)

    def get_relationship(self, secondRel, firstRel, orig_person, other_person):
        common = ""
        if not firstRel:
            if not secondRel:
                return ("", common)
            else:
                return (self.get_direct_ancestor(other_person, secondRel), common)
        elif not secondRel:
            return (self.get_direct_descendant(other_person, firstRel), common)
        elif len(firstRel) == 1:
            if other_person == Person.MALE:
                return (self.get_ancestors_brother(secondRel), common)
            else:
                return (self.get_ancestors_sister(secondRel), common)
        elif len(secondRel) >= len(firstRel):
            return (
                self.get_ancestors_cousin(other_person, secondRel, firstRel),
                common,
            )
        else:
            return (
                self.get_cousins_descendant(other_person, firstRel, secondRel),
                common,
            )

    def get_single_relationship_string(
        self,
        Ga,
        Gb,
        gender_a,
        gender_b,
        reltocommon_a,
        reltocommon_b,
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        return self.get_relationship(reltocommon_a, reltocommon_b, gender_a, gender_b)[
            0
        ]

    def get_sibling_relationship_string(
        self, sib_type, gender_a, gender_b, in_law_a=False, in_law_b=False
    ):
        return self.get_two_way_rel(gender_b, "", "")


if __name__ == "__main__":
    # Test function. Call it as follows from the command line (so as to find
    #        imported modules):
    #    export PYTHONPATH=/path/to/gramps/src
    # python src/plugins/rel/rel_no.py
    # (Above not needed here)

    """TRANSLATORS, copy this if statement at the bottom of your
    rel_xx.py module, and test your work with:
    python src/plugins/rel/rel_xx.py
    """
    from gramps.gen.relationship import test

    RC = RelationshipCalculator()
    test(RC, True)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2010       Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# GTK libraries
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------


# -------------------------------------------------------------------------
#
# SurnamModel
#
# -------------------------------------------------------------------------
class SurnameModel(Gtk.ListStore):
    def __init__(self, surn_list, db):
        # setup model for the treeview
        Gtk.ListStore.__init__(self, str, str, str, str, bool, object)
        for surn in surn_list:
            # fill the liststore
            self.append(
                row=[
                    surn.get_prefix(),
                    surn.get_surname(),
                    surn.get_connector(),
                    str(surn.get_origintype()),
                    surn.get_primary(),
                    surn,
                ]
            )
        self.db = db



# encoding:utf-8
#
# Gramps - a GTK+/GNOME based genealogy program - Records plugin
#
# Copyright (C) 2008-2011 Reinhard Müller
# Copyright (C) 2010      Jakim Friant
# Copyright (C) 2012      Brian G. Matherly
# Copyright (C) 2013-2016 Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Records Report"""

# ------------------------------------------------------------------------
#
# Standard Python modules
#
# ------------------------------------------------------------------------

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.plugins.lib.librecords import (
    RECORDS,
    find_records,
    CALLNAME_DONTUSE,
    CALLNAME_REPLACE,
    CALLNAME_UNDERLINE_ADD,
)
from gramps.gen.plug.docgen import (
    FontStyle,
    ParagraphStyle,
    FONT_SANS_SERIF,
    PARA_ALIGN_CENTER,
    IndexMark,
    INDEX_TYPE_TOC,
)
from gramps.gen.plug.menu import (
    BooleanOption,
    EnumeratedListOption,
    FilterOption,
    NumberOption,
    PersonOption,
    StringOption,
)
from gramps.gen.plug.report import Report
from gramps.gen.plug.report import utils
from gramps.gen.plug.report import MenuReportOptions
from gramps.gen.plug.report import stdoptions
from gramps.gen.lib import Span
from gramps.gen.errors import ReportError
from gramps.gen.proxy import LivingProxyDb, CacheProxyDb


# ------------------------------------------------------------------------
#
# Records Report
#
# ------------------------------------------------------------------------
class RecordsReport(Report):
    """Records Report"""

    def __init__(self, database, options, user):
        """
        This report needs the following parameters (class variables)
        that come in the options class.

        incl_private    - Whether to include private data
        living_people - How to handle living people
        years_past_death - Consider as living this many years after death
        """

        Report.__init__(self, database, options, user)
        menu = options.menu

        self.set_locale(options.menu.get_option_by_name("trans").get_value())

        stdoptions.run_private_data_option(self, menu)
        living_opt = stdoptions.run_living_people_option(self, menu, self._locale)
        self.database = CacheProxyDb(self.database)

        self._lv = menu.get_option_by_name("living_people").get_value()
        for value, description in living_opt.get_items(xml_items=True):
            if value == self._lv:
                living_desc = self._(description)
                break
        self.living_desc = self._("(Living people: %(option_name)s)") % {
            "option_name": living_desc
        }

        filter_option = menu.get_option_by_name("filter")
        self.filter = filter_option.get_filter()

        self.top_size = menu.get_option_by_name("top_size").get_value()
        self.callname = menu.get_option_by_name("callname").get_value()

        self.footer = menu.get_option_by_name("footer").get_value()

        self.include = {}
        for text, varname, default in RECORDS:
            self.include[varname] = menu.get_option_by_name(varname).get_value()

        self._nf = stdoptions.run_name_format_option(self, menu)

    def write_report(self):
        """
        Build the actual report.
        """

        records = find_records(
            self.database,
            self.filter,
            self.top_size,
            self.callname,
            trans_text=self._,
            name_format=self._nf,
            living_mode=self._lv,
            user=self._user,
        )

        self.doc.start_paragraph("REC-Title")
        title = self._("Records")
        mark = IndexMark(title, INDEX_TYPE_TOC, 1)
        self.doc.write_text(title, mark)
        self.doc.end_paragraph()

        self.doc.start_paragraph("REC-Subtitle")
        filter_name = self.filter.get_name(self._locale)
        self.doc.write_text("(%s)" % filter_name)
        self.doc.end_paragraph()
        if self._lv != LivingProxyDb.MODE_INCLUDE_ALL:
            self.doc.start_paragraph("REC-Subtitle")
            self.doc.write_text(self.living_desc)
            self.doc.end_paragraph()

        for text, varname, top in records:
            if not self.include[varname]:
                continue

            self.doc.start_paragraph("REC-Heading")
            self.doc.write_text(self._(text))
            self.doc.end_paragraph()

            last_value = None
            rank = 0
            for number, (sort, value, name, handletype, handle) in enumerate(top):
                mark = None
                if handletype == "Person":
                    person = self.database.get_person_from_handle(handle)
                    mark = utils.get_person_mark(self.database, person)
                elif handletype == "Family":
                    family = self.database.get_family_from_handle(handle)
                    # librecords.py checks that the family has both
                    # a father and a mother and also that each one is
                    # in the filter if any filter was used, so we don't
                    # have to do any similar checking here, it's been done
                    f_handle = family.get_father_handle()
                    dad = self.database.get_person_from_handle(f_handle)
                    f_mark = utils.get_person_mark(self.database, dad)
                    m_handle = family.get_mother_handle()
                    mom = self.database.get_person_from_handle(m_handle)
                    m_mark = utils.get_person_mark(self.database, mom)
                else:
                    raise ReportError(
                        _(
                            "Option '%(opt_name)s' is present "
                            "in %(file)s\n  but is not known to "
                            "the module.  Ignoring..."
                        )
                        % {"opt_name": handletype, "file": "libnarrate.py"}
                    )
                    # since the error is very unlikely I reused the string
                if value != last_value:
                    last_value = value
                    rank = number
                self.doc.start_paragraph("REC-Normal")
                self.doc.write_text(self._("%(number)s. ") % {"number": rank + 1})
                self.doc.write_markup(str(name), name.get_tags(), mark)
                if handletype == "Family":
                    self.doc.write_text("", f_mark)
                    self.doc.write_text("", m_mark)
                if isinstance(value, Span):
                    tvalue = value.get_repr(dlocale=self._locale)
                else:
                    tvalue = value
                self.doc.write_text(" (%s)" % tvalue)
                self.doc.end_paragraph()

        self.doc.start_paragraph("REC-Footer")
        self.doc.write_text(self.footer)
        self.doc.end_paragraph()


# ------------------------------------------------------------------------
#
# Records Report Options
#
# ------------------------------------------------------------------------
class RecordsReportOptions(MenuReportOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, dbase):
        self.__pid = None
        self.__filter = None
        self.__db = dbase
        self._nf = None
        MenuReportOptions.__init__(self, name, dbase)

    def get_subject(self):
        """Return a string that describes the subject of the report."""
        return self.__filter.get_filter().get_name()

    def add_menu_options(self, menu):
        category_name = _("Report Options")

        self.__filter = FilterOption(_("Filter"), 0)
        self.__filter.set_help(_("Determines what people are included in the report."))
        menu.add_option(category_name, "filter", self.__filter)
        self.__filter.connect("value-changed", self.__filter_changed)

        self.__pid = PersonOption(_("Filter Person"))
        self.__pid.set_help(_("The center person for the filter"))
        menu.add_option(category_name, "pid", self.__pid)
        self.__pid.connect("value-changed", self.__update_filters)

        top_size = NumberOption(_("Number of ranks to display"), 3, 1, 100)
        menu.add_option(category_name, "top_size", top_size)

        callname = EnumeratedListOption(_("Use call name"), CALLNAME_DONTUSE)
        callname.set_items(
            [
                (CALLNAME_DONTUSE, _("Don't use call name")),
                (CALLNAME_REPLACE, _("Replace first names with call name")),
                (
                    CALLNAME_UNDERLINE_ADD,
                    _(
                        "Underline call name in first names / "
                        "add call name to first name"
                    ),
                ),
            ]
        )
        menu.add_option(category_name, "callname", callname)

        footer = StringOption(_("Footer text"), "")
        menu.add_option(category_name, "footer", footer)

        category_name = _("Report Options (2)")

        self._nf = stdoptions.add_name_format_option(menu, category_name)
        self._nf.connect("value-changed", self.__update_filters)

        self.__update_filters()

        stdoptions.add_private_data_option(menu, category_name)

        stdoptions.add_living_people_option(menu, category_name)

        stdoptions.add_localization_option(menu, category_name)

        p_count = 0
        for text, varname, default in RECORDS:
            if varname.startswith("person"):
                p_count += 1
        p_half = p_count // 2
        p_idx = 0
        for text, varname, default in RECORDS:
            option = BooleanOption(_(text), default)
            if varname.startswith("person"):
                if p_idx >= p_half:
                    category_name = _("Person 2")
                else:
                    category_name = _("Person 1")
                p_idx += 1
            elif varname.startswith("family"):
                category_name = _("Family")
            menu.add_option(category_name, varname, option)

    def __update_filters(self):
        """
        Update the filter list based on the selected person
        """
        gid = self.__pid.get_value()
        person = self.__db.get_person_from_gramps_id(gid)
        nfv = self._nf.get_value()
        filter_list = utils.get_person_filters(
            person, include_single=False, name_format=nfv
        )
        self.__filter.set_filters(filter_list)

    def __filter_changed(self):
        """
        Handle filter change. If the filter is not specific to a person,
        disable the person option
        """
        filter_value = self.__filter.get_value()
        if filter_value == 0:  # "Entire Database" (as "include_single=False")
            self.__pid.set_available(False)
        else:
            # The other filters need a center person (assume custom ones too)
            self.__pid.set_available(True)

    def make_default_style(self, default_style):
        # Paragraph Styles
        font = FontStyle()
        font.set_type_face(FONT_SANS_SERIF)
        font.set_size(16)
        font.set_bold(True)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_alignment(PARA_ALIGN_CENTER)
        para.set_description(_("The style used for the title."))
        default_style.add_paragraph_style("REC-Title", para)

        font = FontStyle()
        font.set_type_face(FONT_SANS_SERIF)
        font.set_size(12)
        font.set_bold(True)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_alignment(PARA_ALIGN_CENTER)
        para.set_description(_("The style used for the subtitle."))
        default_style.add_paragraph_style("REC-Subtitle", para)

        font = FontStyle()
        font.set_size(12)
        font.set_bold(True)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_top_margin(utils.pt2cm(6))
        para.set_description(_("The style used for the section headers."))
        default_style.add_paragraph_style("REC-Heading", para)

        font = FontStyle()
        font.set_size(10)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_left_margin(0.5)
        para.set_description(_("The basic style used for the text display."))
        default_style.add_paragraph_style("REC-Normal", para)

        font = FontStyle()
        font.set_size(8)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_alignment(PARA_ALIGN_CENTER)
        para.set_top_border(True)
        para.set_top_margin(utils.pt2cm(8))
        para.set_description(_("The style used for the footer."))
        default_style.add_paragraph_style("REC-Footer", para)
